
#include <fat.h>
#include "file.h"

bool
DataFile::init()
{
  PA_WaitForVBL(); PA_WaitForVBL(); PA_WaitForVBL();
  PA_InitFat();
  return true;
}

//-----------------------------------------------------------------

bool
DataFile::write(const std::string &filename)
{
  static std::string HEX_CHARS = "0123456789abcdef";

  std::string full_filename = "fat:/" + filename;
  FILE* testWrite = fopen (full_filename.c_str(), "w");
  if (testWrite)
  {
    for (DataMap::const_iterator map_iter = m_mapData.begin();
	 map_iter != m_mapData.end();
	 ++map_iter)
    {
      unsigned char buffer[BLOCK_SIZE];
      buffer[BLOCK_SIZE-1]='\n';
      for (unsigned int i=0; i < BLOCK_SIZE-1; ++i) { buffer[i] = ' '; }
      fwrite(buffer, 1, BLOCK_SIZE, testWrite);
      for (unsigned int i=0;
	   i < BLOCK_SIZE-1 && i < map_iter->first.size();
	   ++i)
      { buffer[i] = map_iter->first[i]; }
      fwrite(buffer, 1, BLOCK_SIZE, testWrite);

      for (std::vector<u32>::const_iterator vec_iter = map_iter->second.begin();
	   vec_iter != map_iter->second.end();
	   ++vec_iter)
      {
	u32 val = *vec_iter;
	for (unsigned int i=0; i < BLOCK_SIZE-1; ++i)
	{
	  u8 hex = val & 15;
	  buffer[BLOCK_SIZE - i - 2] = HEX_CHARS[hex];
	  val = val >> 4;
	}
	fwrite(buffer, 1, BLOCK_SIZE, testWrite);
      }
    }
    fclose(testWrite);
    return true;
  }
  else
  {
    std::string message = "  Write failed for file " + full_filename;
    PA_OutputText(1, 0, 10, const_cast<char*>(message.c_str()));
    return false;
  }
}

//-----------------------------------------------------------------

bool
DataFile::read(const std::string &filename)
{
  static std::string HEX_CHARS = "0123456789abcdef";

  std::string full_filename = "fat:/" + filename;
  FILE* testWrite = fopen (full_filename.c_str(), "r");
  if (testWrite)
  {
    u8 i=0;
    u8 status=0;
    std::string name("");
    std::vector<u32> data;
    unsigned char buffer[BLOCK_SIZE];
    while (fread(buffer, 1, BLOCK_SIZE, testWrite) > 0)
    {
      if (buffer[0] == ' ')
      {
	// Add to map if need be.
	if (!name.empty()) { m_mapData[name] = data; }
	status = 0;
	name = "";
	data.clear();
      }
      else if (status == 0)
      {
	// read a name.
	for (u8 j=0; j < BLOCK_SIZE-1 && buffer[j] != ' '; ++j)
	{
	  name.push_back(buffer[j]);
	}
	status = 1;
      }
      else if (status == 1)
      {
	// read a number.
	u32 val(0);
	for (u8 j=0; j < BLOCK_SIZE-1; ++j)
	{
	  u32 ind = HEX_CHARS.find(buffer[j]);
	  if (ind < HEX_CHARS.size())
	  {
	    val += (ind << (4 * (BLOCK_SIZE - j - 2)));
	  }
	}
	data.push_back(val);
      }
      i += 1;
    }

    // Don't forget the last item.
    if (!name.empty()) { m_mapData[name] = data; }

    fclose(testWrite);
    return true;
  }
  else { return false; }
}

//-----------------------------------------------------------------

void
DataFile::add(const std::string &name, u32 val)
{
  m_mapData[name] = std::vector<u32>(1, val);
}

//-----------------------------------------------------------------

u32
DataFile::lookup(const std::string &name) const
{
  DataMap::const_iterator iter = m_mapData.find(name);
  if (iter != m_mapData.end() && !iter->second.empty())
  {
    return iter->second[0];
  }
  else { return 0; }
}

//-----------------------------------------------------------------

void
DataFile::addVec(const std::string &name, const std::vector<u32> &vec)
{
  m_mapData[name] = vec;
}

//-----------------------------------------------------------------

std::vector<u32>
DataFile::lookupVec(const std::string &name) const
{
  DataMap::const_iterator iter = m_mapData.find(name);
  if (iter != m_mapData.end() && !iter->second.empty())
  {
    return iter->second;
  }
  else { return std::vector<u32>(); }
}
