//    Kanji Strike - game to learn kanji readings
//    Copyright (C) 2008-2009  Kimberly Dorn

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

//    Contact email: kkdorn@gmail.com

#include <cmath>
#include <sstream>
#include "consts.h"
#include "data.h"
#include "get_chars.h"
#include "menu.h"

//------------------------------------------------------------------

void
PlayerRecordDisplay::show()
{
  for (u8 i=0;
       i < NUM_GRADE_PROGRESS_BAR_SPRITES;
       ++i)
  {
    PA_DualSetSpriteY(SPRITE_BEGIN_GRADE_PROGRESS_BAR + i, 16);
  }

  for (u8 i=0;
       i < NUM_TOTAL_PROGRESS_BAR_SPRITES;
       ++i)
  {
    PA_DualSetSpriteY(SPRITE_BEGIN_TOTAL_PROGRESS_BAR + i, 48);
  }
}

//------------------------------------------------------------------

void
PlayerRecordDisplay::hide()
{
  for (u8 i=0;
       i < NUM_GRADE_PROGRESS_BAR_SPRITES;
       ++i)
  {
    PA_DualSetSpriteY(SPRITE_BEGIN_GRADE_PROGRESS_BAR + i, 384);
  }

  for (u8 i=0;
       i < NUM_TOTAL_PROGRESS_BAR_SPRITES;
       ++i)
  {
    PA_DualSetSpriteY(SPRITE_BEGIN_TOTAL_PROGRESS_BAR + i, 384);
  }

  PA_OutputText(1, 0, 0, "                                ");
  PA_OutputText(1, 0, ROW_GRADE, "                                ");
  PA_OutputText(1, 0, ROW_TOTAL, "                                ");
}

//------------------------------------------------------------------

void
PlayerRecordDisplay::update()
{
  u8 grade = GameData::record.currentGrade();
  u8 grade_perc = GameData::record.gradeCompletionPercentage(grade);
  PA_OutputText(1, 2, ROW_GRADE, "Grade %d", grade+1);
  PA_OutputText(1, 26, ROW_GRADE, "%d %", grade_perc);
  PA_OutputText(1, 14, ROW_GRADE, "%d / %d",
		GameData::record.sizeByStatus(grade, PlayerRecord::Mastered),
		GameData::content.size(grade));

  u16 total(0);
  for (u8 i=0; i <= grade; ++i)
  {
    total += GameData::record.sizeByStatus(i, PlayerRecord::Mastered);
  }
  PA_OutputText(1, 2, ROW_TOTAL, "Total");
  PA_OutputText(1, 26, ROW_TOTAL, "%d %", (total * 100) / GameData::content.size());
  PA_OutputText(1, 14, ROW_TOTAL, "%d / %d",
		total, GameData::content.size());


  u8 scaled_completion = 0.56 * grade_perc;

  for (u8 i=0;
       i < NUM_GRADE_PROGRESS_BAR_SPRITES;
       ++i)
  {
    u8 frame(0);
    if (scaled_completion > (8 * i))
    {
      if (scaled_completion < (8 * (i + 1)))
      {
	frame = scaled_completion - (8 * i);
      }
      else { frame = 8; }
    }
    PA_DualSetSpriteAnim(SPRITE_BEGIN_GRADE_PROGRESS_BAR + i, frame);
  }

  for (u8 i=0; i < NUM_TOTAL_PROGRESS_BAR_SPRITES; ++i)
  {
    u8 frame(0);
    if (i <= grade)
    {
      u8 bar_size(6);
      if (i == 0) { bar_size = 3; }
      else if (i < 6) { bar_size = 5; }
      frame = ((bar_size
		* GameData::record.gradeCompletionPercentage(i))
	       / 100);
    }
    PA_DualSetSpriteAnim(SPRITE_BEGIN_TOTAL_PROGRESS_BAR+i, frame);
  }
}

//------------------------------------------------------------------

void
PlayerRecordDisplay::initSprites()
{
  PA_DualLoadSpritePal(PALETTE_DATA, (void*)data_Pal);

  for (u8 i=0;
       i < NUM_GRADE_PROGRESS_BAR_SPRITES;
       ++i)
  {
    PA_DualCreateSprite(SPRITE_BEGIN_GRADE_PROGRESS_BAR + i,
			(void*)grade_progress_bar_Sprite,
			OBJ_SIZE_32X16,1, PALETTE_DATA,
			(i*32) + 16, 384);
    PA_DualSetSpritePrio(SPRITE_BEGIN_GRADE_PROGRESS_BAR + i, 2);
  }

  for (u8 i=0;
       i < NUM_TOTAL_PROGRESS_BAR_SPRITES;
       ++i)
  {
    const unsigned char* sprite=NULL;
    u8 x=0;
    switch(i)
    {
    case 0:  sprite=grade1_Sprite; x=16; break;
    case 1:  sprite=grade2_Sprite; x=28; break;
    case 2:  sprite=grade34_Sprite; x=46; break;
    case 3:  sprite=grade34_Sprite; x=64; break;
    case 4:  sprite=grade56_Sprite; x=82; break;
    case 5:  sprite=grade56_Sprite; x=100; break;
    case 6:  sprite=grade789_Sprite; x=118; break;
    case 7:  sprite=grade789_Sprite; x=142; break;
    case 8:  sprite=grade789_Sprite; x=166; break;
    case 9:  sprite=gradeab_Sprite; x=190; break;
    case 10: sprite=gradeab_Sprite; x=214; break;
    default: sprite=NULL; x = (i * 32);
    }
    PA_DualCreateSprite(SPRITE_BEGIN_TOTAL_PROGRESS_BAR + i,
			(void*)sprite,
			OBJ_SIZE_32X16,1, PALETTE_DATA,
			x, 384);
    PA_DualSetSpritePrio(SPRITE_BEGIN_TOTAL_PROGRESS_BAR + i, 2);
  }
}

//------------------------------------------------------------------

GameplayMenu::GameplayMenu()
  : m_triggerLeft("TITLE", true),
    m_triggerRight("DATA", false),
    m_nExit(Title)
{
  std::vector<std::string> speed_options;
  speed_options.push_back("slow");
  speed_options.push_back("medium");
  speed_options.push_back("fast");
  m_interface.addSelectItem(Speed, "Speed:", speed_options);

  std::vector<std::string> meteor_options;
  meteor_options.push_back("4");
  meteor_options.push_back("5");
  meteor_options.push_back("6");
  m_interface.addSelectItem(Meteors, "Blocks:", meteor_options);

  m_interface.setStyle(MenuInterface::Style(MenuInterface::Style::Centered,
					    0, 24, 2, 4));
  m_interface.addTextItem(Go, "GO");
  m_interface.hide();
}

//------------------------------------------------------------------

bool
GameplayMenu::loop()
{
  bool result(true);
  m_interface.handleStandardInput();
  m_triggerLeft.handleStandardInput();
  m_triggerRight.handleStandardInput();
  if (m_interface.clickEvent() && m_interface.currentIndex() == Go)
  {
    m_nExit = Game; result = false;
  }
  else if (m_triggerLeft.clickEvent()) { m_nExit = Title; result = false; }
  else if (m_triggerRight.clickEvent()) { m_nExit = Data; result = false; }
  PA_WaitForVBL();
  return result;
}

//------------------------------------------------------------------

u32
GameplayMenu::getSpeed()
{
  u8 i = m_interface.index(Speed);
  switch(i)
  {
  case 0: return 200; break;
  case 1: return 140; break;
  default: return 80;
  }
}

//------------------------------------------------------------------

u8
GameplayMenu::getMeteors()
{
  u8 i = m_interface.index(Meteors);
  switch(i)
  {
  case 0: return 4; break;
  case 1: return 5; break;
  default: return 6;
  }
}

//------------------------------------------------------------------

void
GameplayMenu::run()
{
  bool keep_going(true);

  PA_DualSetBgColor(PA_RGB(7, 7, 7));
  m_interface.show();
  m_triggerLeft.show();
  m_triggerRight.show();
  PlayerRecordDisplay::update();
  PlayerRecordDisplay::show();
  while (keep_going)
  {
    keep_going = loop();
  }
  m_interface.hide();
  m_triggerLeft.hide();
  m_triggerRight.hide();
  PlayerRecordDisplay::hide();
  PA_DualSetBgColor(PA_RGB(0, 0, 0));
}

//------------------------------------------------------------------

MenuPage
GameplayMenu::exit() const { return m_nExit; }

//------------------------------------------------------------------

DataMenu::DataMenu()
  : m_triggerLeft("TITLE", true),
    m_triggerRight("GAME", false)
{
  m_interface.setStyle(MenuInterface::Style(MenuInterface::Style::Centered,
					    0, 24, 2, 4));
  m_interface.addTextItem(ViewMastered, "View mastered");
  m_interface.addTextItem(Delete, "Delete this file");
  m_interface.hide();
}

//------------------------------------------------------------------

bool
DataMenu::loop()
{
  bool result(true);
  m_interface.handleStandardInput();
  m_triggerLeft.handleStandardInput();
  m_triggerRight.handleStandardInput();
  //  u8 ce = m_interface.clickEvent();
  if (m_interface.clickEvent())
  {
    if (m_interface.currentIndex() == Delete)
    {
      m_nExit = DeleteConfirm;
      result = false;
    }
    else if (m_interface.currentIndex() == ViewMastered)
    {
      m_nExit = ViewData;
      result = false;
    }
  }
  else if (m_triggerLeft.clickEvent()) { m_nExit = Title; result = false; }
  else if (m_triggerRight.clickEvent())
  {
    m_nExit = Gameplay;
    result = false;
  }
  PA_WaitForVBL();
  return result;
}

//------------------------------------------------------------------

void
DataMenu::run()
{
  bool keep_going(true);

  PA_DualSetBgColor(PA_RGB(7, 7, 7));
  m_interface.show();
  m_triggerLeft.show();
  m_triggerRight.show();
  PlayerRecordDisplay::update();
  PlayerRecordDisplay::show();
  while (keep_going)
  {
    keep_going = loop();
  }
  m_interface.hide();
  m_triggerLeft.hide();
  m_triggerRight.hide();
  PlayerRecordDisplay::hide();
  PA_DualSetBgColor(PA_RGB(0, 0, 0));
}

//------------------------------------------------------------------

MenuPage DataMenu::exit() const { return m_nExit; }

//------------------------------------------------------------------

DeleteConfirmMenu::DeleteConfirmMenu()
{
  m_interface.setStyle(MenuInterface::Style(MenuInterface::Style::Centered,
					    0, 24, 2, 4));
  m_interface.addTextItem(DeleteYes, "Delete");
  m_interface.addTextItem(DeleteNo, "Cancel delete");
  m_interface.hide();
}

//------------------------------------------------------------------

bool
DeleteConfirmMenu::loop()
{
  bool result(true);
  m_interface.handleStandardInput();
  if (m_interface.clickEvent())
  {
    if (m_interface.currentIndex() == DeleteYes)
    {
      // delete file.
      m_nExit = DeleteFile;
      result = false;
    }
    else if (m_interface.currentIndex() == DeleteNo)
    {
      m_nExit = Data;
      result = false;
    }
  }
  PA_WaitForVBL();
  return result;
}


//------------------------------------------------------------------

void
DeleteConfirmMenu::run()
{
  bool keep_going(true);

  PA_DualSetBgColor(PA_RGB(7, 7, 7));
  m_interface.show();
  PlayerRecordDisplay::update();
  PlayerRecordDisplay::show();
  while (keep_going)
  {
    keep_going = loop();
  }
  m_interface.hide();
  PlayerRecordDisplay::hide();
  PA_DualSetBgColor(PA_RGB(0, 0, 0));
}

//------------------------------------------------------------------

MenuPage DeleteConfirmMenu::exit() const { return m_nExit; }

//------------------------------------------------------------------

TitleMenu::TitleMenu()
{
  m_interface.hide();
}

//------------------------------------------------------------------

bool
TitleMenu::loop()
{
  bool result(true);
  m_interface.handleStandardInput();
  if (m_interface.clickEvent())
  {
    result = false;
    if (m_interface.currentIndex() < 3)
    {
      m_nExit = LoadGame;
    }
    else
    {
      m_nExit = Guide;
    }
  }
  PA_WaitForVBL();
  return result;
}

//------------------------------------------------------------------

void
TitleMenu::init()
{
  bool read_ok(true);
  m_interface = MenuInterface();
  m_interface.setStyle(MenuInterface::Style(MenuInterface::Style::Centered,
					    0, 24, 2, 4));
  for (u8 i=0; i < 3; ++i)
  {
    PlayerRecord game;
    read_ok = read_ok && GameData::readNumberedFile(i+1, game);

    u16 total(0);
    for (u8 j=0; j <= GameData::content.numGrades(); ++j)
    {
      total += game.sizeByStatus(j, PlayerRecord::Mastered);
    }
    
    u16 grade = game.currentGrade();
    u16 grade_perc = game.gradeCompletionPercentage(grade);
    u16 perc_complete = (100 * total) / GameData::content.size();

    std::stringstream item_ss;
    item_ss << "Game " << i+1 << "     "
	    << "Grade " << grade + 1 << "  " << grade_perc << "%";
    m_interface.addTextItem(i, item_ss.str());
  }
  m_interface.addTextItem(3, "How it works");

  if (!read_ok)
  {
    //    PA_OutputText(1, 0, 14, "ERROR: Save file read failed.");
  }
}

//------------------------------------------------------------------

void
TitleMenu::run()
{
  bool keep_going(true);

  PA_DualSetBgColor(PA_RGB(7, 7, 7));
  m_interface.hide();
  init();
  m_interface.show();
  PA_OutputText(1, 6, 10, "KANJI STRIKE  v1.0");
  PA_OutputText(1, 0, 16, "Make sure your root directory");
  PA_OutputText(1, 0, 17, "contains a folder named 'ks'");
  PA_OutputText(1, 0, 18, "to store save files.");

  while (keep_going)
  {
    keep_going = loop();
  }
  m_interface.hide();
  PA_OutputText(1, 6, 10, "                  ");
  PA_OutputText(1, 0, 14, "                                ");
  PA_OutputText(1, 0, 16, "                                ");
  PA_OutputText(1, 0, 17, "                                ");
  PA_OutputText(1, 0, 18, "                                ");
  PA_DualSetBgColor(PA_RGB(0, 0, 0));
}

//------------------------------------------------------------------

MenuPage
TitleMenu::exit() const { return m_nExit; }

//------------------------------------------------------------------

u8
TitleMenu::fileIndex() const { return m_interface.currentIndex()+1; }

//------------------------------------------------------------------

GuideMenu::GuideMenu()
{
  m_interface.addTextItem(GuideMenuGameplay, "Gameplay");
  m_interface.addTextItem(GuideMenuData, "Data");
  m_interface.addTextItem(GuideMenuControls, "Controls");
  m_interface.addTextItem(GuideMenuAcknowledgements, "Acknowledgements");
  m_interface.addTextItem(GuideMenuBackToTitle, "Back to title screen");
    m_interface.setStyle(MenuInterface::Style(MenuInterface::Style::Centered,
					      0, 24, 2, 4));
}

//------------------------------------------------------------------

void
GuideMenu::run()
{
  bool keep_going(true);
  PA_DualSetBgColor(PA_RGB(7, 7, 7));
  m_interface.show();
  while (keep_going)
  {
    m_interface.handleStandardInput();
    if (m_interface.clickEvent())
    {
      keep_going = false;
      if (m_interface.currentIndex() == GuideMenuGameplay)
      {
	m_nExit = GuideGameplay;
      }
      else if (m_interface.currentIndex() == GuideMenuData)
      {
	m_nExit = GuideData;
      }
      else if (m_interface.currentIndex() == GuideMenuControls)
      {
	m_nExit = GuideControls;
      }
      else if (m_interface.currentIndex() == GuideMenuAcknowledgements)
      {
	m_nExit = GuideAcknowledgements;
      }
      else { m_nExit = Title; }
    }
    PA_WaitForVBL();
  }
  m_interface.hide();
  PA_DualSetBgColor(PA_RGB(0, 0, 0));
}

//------------------------------------------------------------------

MenuPage GuideMenu::exit() const { return m_nExit; }

//------------------------------------------------------------------

TextPage::TextPage()
  : m_nLine(0)
{}

//------------------------------------------------------------------

void
TextPage::addTextItem(u8 indent, const std::string &s)
{
  TextItem ti(m_nLine, indent, s);
  m_vecTextItems.push_back(ti);
  m_nLine += ti.numRows();

  // change menu style.
  if (m_nLine > 24)
  {
    m_interface.setStyle(MenuInterface::Style(MenuInterface::Style::Centered,
					      m_nLine - 24, 24, 2, 4));
  }
}

//------------------------------------------------------------------

void TextPage::addBlankLines(u8 lines) { m_nLine += lines; }

//------------------------------------------------------------------

void
TextPage::setBackText(const std::string &s)
{
  m_interface.clear();
  m_interface.addTextItem(s);
}

//------------------------------------------------------------------

void
TextPage::run()
{
  bool keep_going(true);

  PA_DualSetBgColor(PA_RGB(7, 7, 7));

  for (std::vector<TextItem>::iterator iter = m_vecTextItems.begin();
       iter != m_vecTextItems.end();
       ++iter)
  {
    iter->show();
  }
  m_interface.show();

  while (keep_going)
  {
    m_interface.handleStandardInput();
    keep_going = !m_interface.clickEvent();
  }

  m_interface.hide();
  for (std::vector<TextItem>::iterator iter = m_vecTextItems.begin();
       iter != m_vecTextItems.end();
       ++iter)
  {
    iter->hide();
  }

  PA_DualSetBgColor(PA_RGB(0, 0, 0));
}

//------------------------------------------------------------------

DataViewer::Entry::Entry(const std::set<u16> &tr,
			 const u8 &max_rows_per_page,
			 const u8 &max_kana_per_page)
  : m_nUnicode(0), m_setTopRows(tr)
{
  if (!m_setTopRows.empty())
  {
    m_nUnicode = GameData::content.getRow(*(m_setTopRows.begin())).unicode();
  }
  else { return; }

  u8 num_header_kana(0);
  for (std::set<u16>::const_iterator iter = m_setTopRows.begin();
       iter != m_setTopRows.end();
       ++iter)
  {
    num_header_kana += GameData::content.getRow(*iter).reading().size();
  }

  u8 max_kana = max_kana_per_page - num_header_kana;
  u8 kana_used(0);
  u8 rows_used(0);
  ReadingVec page;
  std::pair<u16, u16> range
    = GameData::content.getCharRange(*(m_setTopRows.begin()));  
  for (u16 i=range.first; i < range.second; ++i)
  {
    if (m_setTopRows.find(i) == m_setTopRows.end())
    {
      Reading r("??");
      if (GameData::record.state(i) == PlayerRecord::Mastered)
      {
	r = GameData::content.getRow(i).reading();
      }
      if (kana_used + r.size() > max_kana 
	  || rows_used == max_rows_per_page)
      {
	m_vecPages.push_back(page);
	page.clear();
	kana_used = 0;
	rows_used = 0;
      }
      page.push_back(r);
      kana_used += r.size();
      rows_used += 1;
    }
  }
  if (!page.empty()) { m_vecPages.push_back(page); }
}

//------------------------------------------------------------------

u8 DataViewer::Entry::pageCount() const { return m_vecPages.size(); }

//------------------------------------------------------------------

std::vector<Reading>
DataViewer::Entry::headerReadings() const
{
  std::vector<Reading> res;

  for (std::set<u16>::const_iterator iter = m_setTopRows.begin();
       iter != m_setTopRows.end();
       ++iter)
  {
    res.push_back(GameData::content.getRow(*iter).reading());
  }

  return res;
}

//------------------------------------------------------------------

std::vector<Reading>
DataViewer::Entry::otherReadings(u8 page) const
{
  std::vector<Reading> res;
  if (page >= m_vecPages.size()) { return res; }
  else { return m_vecPages[page]; }
}

//------------------------------------------------------------------

u8
DataViewer::Entry::numRows(u8 page) const
{
  if (page >= m_vecPages.size()) { return 0; }

  return m_setTopRows.size() + m_vecPages[page].size();
}

//------------------------------------------------------------------

u8
DataViewer::Entry::numKana(u8 page) const
{
  if (page >= m_vecPages.size()) { return 0; }

  std::vector<Reading> headers = headerReadings();
  std::vector<Reading> others = otherReadings(page);

  u8 res(0);
  for (std::vector<Reading>::const_iterator iter = headers.begin();
       iter != headers.end();
       ++iter)
  {
    res += iter->size();
  }

  for (std::vector<Reading>::const_iterator iter = others.begin();
       iter != others.end();
       ++iter)
  {
    res += iter->size();
  }

  return res;
}

//------------------------------------------------------------------

u8 DataViewer::Entry::numHeaders() const { return m_setTopRows.size(); }

//------------------------------------------------------------------

u16 DataViewer::Entry::unicode() const { return m_nUnicode; }

//------------------------------------------------------------------

DataViewer::DataViewer()
  : m_buttonLeft(SPRITE_BROWSER_NAV_LEFT_BACKGROUND,
		 64, 64,
		 SPRITE_BROWSER_NAV_LEFT_ARROWS,
		 0, 320, KeyLeft, false,
		 16, 16),
    m_buttonRight(SPRITE_BROWSER_NAV_RIGHT_BACKGROUND,
		  64, 64,
		  SPRITE_BROWSER_NAV_RIGHT_ARROWS,
		  192, 320, KeyRight, false,
		  16, 16),
    m_buttonBack(SPRITE_BROWSER_BACK_BACKGROUND,
		 64, 32,
		 SPRITE_BROWSER_BACK_CONTENT,
		 192, 272, KeyB, false,
		 16, 8),
    m_nNumRows(0),
    m_nKanaSpritesUsed(0),
    m_nKanjiSpritesUsed(0),
    m_nIndex(0),
    m_nNumOnPage(0),
    m_nPageWithinRow(0)
{
  for (u8 i=0; i < NUM_DATA_VIEWER_MENU_OPTIONS; ++i)
  {
    Button b(SPRITE_BEGIN_BROWSER_MENU_BACKGROUND+i,
	     32, 32,
	     SPRITE_BEGIN_BROWSER_MENU_CONTENT+i,
	     72 + (40 * i), 352, KeyNone, true, 0, 0);
    b.setContentSprite(SPRITE_BEGIN_BROWSER_MENU_CONTENT+i, i, 0, 0);
    m_buttonsMenu.addButton(b);
  }
  m_buttonsMenu.hide();
}

//------------------------------------------------------------------

DataViewer::Entry
DataViewer::entry(u16 index) const
{
  if (index >= m_nNumRows)
  {
    return Entry(std::set<u16>(), 19, NUM_KANA);
  }

  u8 consecutive_rows(1);
  std::set<u16> top_rows;
  top_rows.insert(m_vecRows[index]);

  bool done(false);
  while (index > 0 && !done)
  {
    if (row(index).unicode() == row(index-1).unicode()) { index -= 1; }
    else { done = true; }
  }

  while (index + consecutive_rows < m_nNumRows
	 && (row(index).unicode()
	     == row(index + consecutive_rows).unicode()))
  {
    top_rows.insert(m_vecRows[index + consecutive_rows]);
    consecutive_rows += 1;
  }

  // make 19 a const.
  return Entry(top_rows, 19, NUM_KANA);
}

//------------------------------------------------------------------

u16
DataViewer::beginUnicodeRange(u16 index) const
{
  u16 res(index);
  bool done(false);
  while (res > 0 && !done)
  {
    if (row(res).unicode() == row(res-1).unicode()) { res -= 1; }
    else { done = true; }
  }
  return res;
}

//------------------------------------------------------------------

u16
DataViewer::endUnicodeRange(u16 index)
{
  u16 res(index+1);
  while (res < m_nNumRows
	 && row(index).unicode() == row(res).unicode())
  {
    res += 1;
  }
  return res;
}

//------------------------------------------------------------------

u16
DataViewer::firstKana(u16 index)
{
  ContentRow cr = row(index);
  u16 res(Reading::BAD_CHAR);
  u8 i(0);

  while (i < cr.reading().size()
	 && (res == Reading::BAD_CHAR || res == Reading::SEPARATOR))
  {
    res = cr.reading()[i];
    if (res >= 128) { res -= 96; }
    ++i;
  }

  return res;
}

//------------------------------------------------------------------

u16
DataViewer::beginKanaRange(u16 index)
{
  u16 res(index);
  while (res > 0 && firstKana(res) == firstKana(index))
  {
    res -= 1;
  }

  if (firstKana(res) != firstKana(index)) { ++res; }

  return res;
}

//------------------------------------------------------------------

u16
DataViewer::previousKanaIndex()
{
  u16 res = beginKanaRange(m_nIndex);
  if (res)
  {
    res = beginKanaRange(res - 1);
  }

  return res;
}

//------------------------------------------------------------------

u16
DataViewer::nextKanaIndex()
{
  u16 res(BAD_INDEX);
  u16 index(m_nIndex + 1);
  while (index < m_nNumRows && res == BAD_INDEX)
  {
    if (firstKana(m_nIndex) != firstKana(index))
    {
      res = index;
    }
    else { index += 1; }
  }

  return res;
}

//------------------------------------------------------------------

ContentRow
DataViewer::row(u16 index) const
{
  return GameData::content.getRow(m_vecRows[index]);
}

//------------------------------------------------------------------

u16
DataViewer::unicode(u16 index)
{
  return GameData::content.getRow(m_vecRows[index]).unicode();
}

//------------------------------------------------------------------

u8
DataViewer::pageLength(u16 index) const
{
  s32 y(0);
  u8 res(0);
  u16 current_index(index);
  bool done(false);
  u8 total_kanji(0);
  u8 total_kana(0);

  while (!done)
  {
    Entry e = entry(current_index);

    if (e.pageCount() <= 1)
    {
      y += 16 * e.numRows(0);
      total_kanji += 1;
      total_kana += e.numKana(0);
    }

    // if over the limit, set done.
    // else increment current_index and res.
    if (e.pageCount() > 1
	|| y > MAX_Y 
	|| total_kanji > NUM_KANJI
	|| total_kana > NUM_KANA)
    {
      done = true;
    }
    else
    {
      current_index += e.numHeaders();
      res += e.numHeaders();
      done = (current_index >= m_nNumRows);
    }
  }
  return res;
}

//------------------------------------------------------------------

u8
DataViewer::pageCount(u16 index) const
{
  u16 row_index(m_vecRows[index]);
  std::pair<u16, u16> range = GameData::content.getCharRange(row_index);
  return 0;
}

//------------------------------------------------------------------

u16
DataViewer::previousPageTop(const u16 &index) const
{
  if (index == 0) { return 0; }

  u16 res(index-1);
  bool over(false);
  u8 line(0);
  while (res > 0 && !over)
  {
    u8 pl = pageLength(res);
    line += 1;
    if (res + pl <= index)
    {
      over = true;
    }
    else
    {
      res -= 1;
    }
  }
  res = beginUnicodeRange(res + 1);
  //  res += 1;
  if (res >= index) { res = index - 1; }

  return res;
}

//------------------------------------------------------------------

bool
DataViewer::drawReading(u16 content_row_index, s32 x, s32 y)
{
  Reading r = GameData::content.getRow(content_row_index).reading();
  return drawReading(r, x, y);
}

//------------------------------------------------------------------

bool
DataViewer::drawReading(const Reading &r, s32 x, s32 y)
{
  if (m_nKanaSpritesUsed + r.size() > NUM_KANA)
  {
    return false;
  }
  else if (y >= MAX_Y) { return false; }

  for (u8 j=0; j < r.size(); ++j)
  {
    PA_DualUpdateSpriteGfx(SPRITE_BEGIN_KANA
			   + m_nKanaSpritesUsed + j,
			   (void *)getKana(r[j]));
    PA_DualSetSpriteXY(SPRITE_BEGIN_KANA + m_nKanaSpritesUsed + j,
		       x + (16 * j),
		       y);
  }
  m_nKanaSpritesUsed += r.size();
  return true;
}

//------------------------------------------------------------------

void
DataViewer::drawEntry(const Entry &e, u8 page, u8 start_row)
{
  if (m_nKanjiSpritesUsed >= NUM_KANJI) { return; }
  if (start_row + e.numRows(page) >= KANA_ROWS) { return; }
  if (m_nKanaSpritesUsed + e.numKana(page) >= NUM_KANA) { return; }

  PA_DualUpdateSpriteGfx(SPRITE_BEGIN_KANJI+m_nKanjiSpritesUsed,
			 (void *)getKanji(e.unicode()));
  PA_DualSetSpriteXY(SPRITE_BEGIN_KANJI+m_nKanjiSpritesUsed,
		     KANJI_X, start_row * 16);
  m_nKanjiSpritesUsed += 1;

  s32 current_y = start_row * 16;

  ReadingVec headers = e.headerReadings();
  for (ReadingVec::const_iterator iter = headers.begin();
       iter != headers.end();
       ++iter)
  {
    drawReading(*iter, KANA_X, current_y);
    current_y += 16;
  }

  ReadingVec others = e.otherReadings(page);
  for (ReadingVec::const_iterator iter = others.begin();
       iter != others.end();
       ++iter)
  {
    drawReading(*iter, KANA_X + 16, current_y);
    current_y += 16;
  }
}

//------------------------------------------------------------------

void
DataViewer::clearPage()
{
  for (u8 i=0; i < NUM_KANJI; ++i)
  {
    PA_DualSetSpriteXY(SPRITE_BEGIN_KANJI+i, 0, 384);
  }

  for (u8 i=0; i < NUM_KANA; ++i)
  {
    PA_DualSetSpriteXY(SPRITE_BEGIN_KANA+i, 0, 384);
  }

  m_nKanjiSpritesUsed = 0;
  m_nKanaSpritesUsed = 0;
  m_nNumOnPage = 0;

  PA_OutputText(0, 10, 17, "                                ");
  PA_OutputText(0, 10, 18, "                                ");
}

//------------------------------------------------------------------

void
DataViewer::drawPage()
{
  if (m_nIndex >= m_nNumRows) { return; }

  clearPage();

  s32 y=0;

  u16 current_row = m_nIndex;
  u16 current_row_end = endUnicodeRange(m_nIndex);
  std::pair<u16, u16> range = GameData::content.getCharRange(m_vecRows[current_row]);

  bool done(false);
  u8 screen_row(0);
  Entry e = entry(current_row);
  drawEntry(e, m_nPageWithinRow, screen_row);
  m_nNumOnPage += e.numHeaders();
  screen_row += (e.numRows(m_nPageWithinRow) + 2);
  if (screen_row == 10 || screen_row == 11) { screen_row = 12; }
  if (e.pageCount() > 1) { done = true;}

  while (!done)
  {
    current_row = current_row_end;
    current_row_end = endUnicodeRange(current_row);

    if (current_row < m_nNumRows)
    {
      Entry e = entry(current_row);
      if (e.pageCount() < 2 && screen_row + e.numRows(0) < KANA_ROWS)
      {
	drawEntry(e, 0, screen_row);
	m_nNumOnPage += e.numHeaders();
	screen_row += (e.numRows(0) + 2);
	if (screen_row == 10 || screen_row == 11) { screen_row = 12; }
      }
      else { done = true; }
    }
    else { done = true;; }
  }

  if (!m_nNumOnPage) { m_nNumOnPage = 1; }

  PA_OutputText(0, 10, 17, "%d - %d",
		m_nIndex+1, m_nIndex+m_nNumOnPage);
  PA_OutputText(0, 10, 18, " of %d", m_nNumRows);
}

//------------------------------------------------------------------

void
DataViewer::updateKanaButtons()
{
  u16 pki = previousKanaIndex();
  u16 nki = nextKanaIndex();

  if (pki != BAD_INDEX)
  {

    PA_DualUpdateSpriteGfx(SPRITE_BROWSER_NAV_LEFT_KANA,
			   (void *)getKana(firstKana(pki)));
  }

  if (nki != BAD_INDEX)
  {
    PA_DualUpdateSpriteGfx(SPRITE_BROWSER_NAV_RIGHT_KANA,
			   (void *)getKana(firstKana(nki)));
  }
}

//------------------------------------------------------------------

bool
DataViewer::loop()
{
  m_buttonBack.handleStandardInput();
  if (m_buttonBack.checkClick())
  { return false; }

  if (m_nIndex == 0) { m_buttonLeft.hide(); }
  else { m_buttonLeft.show(); }

  if (m_nIndex + m_nNumOnPage >= m_nNumRows) { m_buttonRight.hide(); }
  else { m_buttonRight.show(); }

  bool menu_click = m_buttonsMenu.handleStylusInput();
  if (!menu_click) { menu_click = m_buttonsMenu.handleTriggerInput(); }
  if (menu_click)
  {
    // change nav buttons if need be.
    if (m_buttonsMenu.index() == 0)
    {
      m_buttonLeft.setContentSprite(SPRITE_BROWSER_NAV_LEFT_ARROWS,
				    0, 16, 16);
      m_buttonRight.setContentSprite(SPRITE_BROWSER_NAV_RIGHT_ARROWS,
				     0, 16, 16);
    }
    else if (m_buttonsMenu.index() == 1)
    {
      updateKanaButtons();
      m_buttonLeft.setContentSprite(SPRITE_BROWSER_NAV_LEFT_KANA,
				    0, 24, 24);
      m_buttonRight.setContentSprite(SPRITE_BROWSER_NAV_RIGHT_KANA,
				     0, 24, 24);
      updateKanaButtons();
    }
    else if (m_buttonsMenu.index() == 2)
    {
      m_buttonLeft.setContentSprite(SPRITE_BROWSER_NAV_LEFT_ARROWS,
				    1, 16, 16);
      m_buttonRight.setContentSprite(SPRITE_BROWSER_NAV_RIGHT_ARROWS,
				     1, 16, 16);
    }
  }
  else
  {
    if (!m_buttonLeft.busy() && !m_buttonRight.busy())
    {
      if (m_buttonLeft.checkClick())
      {
	// page left.
	if (m_buttonsMenu.index() == 2)
	{
	  m_nIndex = 0;
	  drawPage();
	}
	else if (m_buttonsMenu.index() == 1)
	{
	  if (previousKanaIndex() != BAD_INDEX)
	  {
	    m_nIndex = previousKanaIndex();
	    updateKanaButtons();
	    drawPage();
	  }
	}
	else
	{
	  if (m_nPageWithinRow == 0)
	  {
	    // if this index has more than one page,
	    // set m_nPageWithinRow to the max.
	    m_nIndex = previousPageTop(m_nIndex);
	    Entry e = entry(m_nIndex);
	    if (e.pageCount() > 1)
	    {
	      m_nPageWithinRow = e.pageCount() - 1;
	    }
	  }
	  else { m_nPageWithinRow -= 1; }
	  drawPage();
	}
      }
      else if (m_buttonRight.checkClick())
      {
	// page right.
	if (m_buttonsMenu.index() == 2)
	{
	  m_nIndex = previousPageTop(m_nNumRows);
	  drawPage();
	}
	else if (m_buttonsMenu.index() == 1)
	{
	  if (nextKanaIndex() != BAD_INDEX)
	  {
	    m_nIndex = nextKanaIndex();
	    updateKanaButtons();
	    drawPage();
	  }
	}
	else if (m_nIndex + m_nNumOnPage < m_nNumRows)
	{
	  // if this index has more than one page, increment page.
	  Entry e = entry(m_nIndex);
	  if (m_nPageWithinRow + 1 < e.pageCount())
	  {
	    m_nPageWithinRow += 1;
	  }
	  else
	  {
	    m_nPageWithinRow = 0;
	    m_nIndex += m_nNumOnPage;
	  }
	  drawPage();
	}
      }

      m_buttonLeft.handleStandardInput();
      m_buttonRight.handleStandardInput();
    }
  }

  PA_WaitForVBL();
  return true;
}

//------------------------------------------------------------------

void
DataViewer::run()
{
  bool keep_going(true);
  PA_DualSetBgColor(PA_RGB(7, 7, 7));

  m_nIndex = 0;
  m_vecRows = GameData::record.getReadingSortedRows();
  m_nNumRows = m_vecRows.size();

  m_buttonLeft.show();
  m_buttonRight.show();
  m_buttonBack.show();
  m_buttonsMenu.show();
  drawPage();
  while (keep_going) { keep_going = loop(); }
  clearPage();
  m_buttonLeft.hide();
  m_buttonRight.hide();
  m_buttonBack.hide();
  m_buttonsMenu.hide();
}

//------------------------------------------------------------------

void
DataViewer::initSprites()
{
  PA_DualCreateSprite(SPRITE_BROWSER_NAV_LEFT_BACKGROUND,
		      (void*)button_back64_Sprite,
		      OBJ_SIZE_64X64, 1, PALETTE_MENU,
		      0, 384);
  PA_DualSetSpritePrio(SPRITE_BROWSER_NAV_LEFT_BACKGROUND, 2);

  PA_DualCreateSprite(SPRITE_BROWSER_NAV_RIGHT_BACKGROUND,
		      (void*)button_back64_Sprite,
		      OBJ_SIZE_64X64, 1, PALETTE_MENU,
		      0, 384);
  PA_DualSetSpritePrio(SPRITE_BROWSER_NAV_RIGHT_BACKGROUND, 2);

  PA_DualCreateSprite(SPRITE_BROWSER_NAV_LEFT_ARROWS,
		      (void*)browser_left_nav_Sprite,
		      OBJ_SIZE_32X32, 1, PALETTE_MENU,
		      0, 384);
  PA_DualSetSpritePrio(SPRITE_BROWSER_NAV_LEFT_ARROWS, 1);

  PA_DualCreateSprite(SPRITE_BROWSER_NAV_RIGHT_ARROWS,
		      (void*)browser_right_nav_Sprite,
		      OBJ_SIZE_32X32, 1, PALETTE_MENU,
		      0, 384);
  PA_DualSetSpritePrio(SPRITE_BROWSER_NAV_RIGHT_ARROWS, 1);

  PA_DualCreateSprite(SPRITE_BROWSER_NAV_LEFT_KANA,
		      NULL,
		      OBJ_SIZE_16X16, 0, PALETTE_CHARS,
		      0, 384);
  PA_DualSetSpritePrio(SPRITE_BROWSER_NAV_LEFT_ARROWS, 1);

  PA_DualCreateSprite(SPRITE_BROWSER_NAV_RIGHT_KANA,
		      NULL,
		      OBJ_SIZE_16X16, 0, PALETTE_CHARS,
		      0, 384);
  PA_DualSetSpritePrio(SPRITE_BROWSER_NAV_RIGHT_ARROWS, 1);

  PA_DualCreateSprite(SPRITE_BROWSER_BACK_BACKGROUND,
		      (void*)button_back64x32_Sprite,
		      OBJ_SIZE_64X32, 1, PALETTE_MENU,
		      0, 384);
  PA_DualSetSpritePrio(SPRITE_BROWSER_BACK_BACKGROUND, 2);

  PA_DualCreateSprite(SPRITE_BROWSER_BACK_CONTENT,
		      (void*)browser_back_Sprite,
		      OBJ_SIZE_32X16, 1, PALETTE_MENU,
		      0, 384);
  PA_DualSetSpritePrio(SPRITE_BROWSER_BACK_CONTENT, 1);

  for (u8 i=0; i < NUM_DATA_VIEWER_MENU_OPTIONS; ++i)
  {
    PA_DualCreateSprite(SPRITE_BEGIN_BROWSER_MENU_BACKGROUND+i,
			(void*)button_back_Sprite,
			OBJ_SIZE_32X32, 1, PALETTE_MENU,
			0, 384);
    PA_DualSetSpritePrio(SPRITE_BEGIN_BROWSER_MENU_BACKGROUND+i, 2);

    PA_DualCreateSprite(SPRITE_BEGIN_BROWSER_MENU_CONTENT+i,
			(void*)browser_menu_Sprite,
			OBJ_SIZE_32X32, 1, PALETTE_MENU,
			0, 384);
    PA_DualSetSpritePrio(SPRITE_BEGIN_BROWSER_MENU_CONTENT+i, 1);
  }
}

//------------------------------------------------------------------

MenuLoop::MenuLoop() : m_bFirstLoad(true)
{
  std::string gp1 = "Blocks labelled with kanji fall toward the bottom"
    " of the screen.  The point of the game is to shoot down the blocks"
    " with missiles labelled with pronunciations.";
  std::string gp2 = " If a block reaches the bottom of the screen, it explodes"
    " and causes the player damage.  When the player's life bar is empty,"
    " the game ends.";
  std::string gp3 = "In the gameplay menu, the player can control"
    " how many blocks there are and how fast they fall.";

  std::string ack_kcd = "Kanji Strike is based on the game"
    " Kanji City Defender.";
  std::string ack_kanjidic = "This game uses data extracted"
    " from the KANJIDIC dictionary file. Its documentation can be found"
    " at http://www.csse.monash.edu.au/~jwb/kanjidic_doc.html."
    "  This file is the property of the Electronic Dictionary Research"
    " and Development Group, and is used in conformance with the Group's"
    " licence.";
  std::string ack_font = "This game uses the P Gothic font"
    " from the Ume-Font project (http://sourceforge.jp/projects/ume-font)."
    " The font is the property of the Electronic Font Open Laboratory,"
    " and is used in conformance with their license.";
  std::string ack_palib = "This version of Kanji Strike was implemented"
    " using the PALib library (http://www.palib.info/).";
  std::string ack_ks = "For more information about the game,"
    " please visit kanjistrike.sourceforge.net.";

  std::string data1 = "The material is organized as pairs of a kanji"
    " and a pronunciation.  It is divided into 11 grades by kanji.";
  std::string data2 = "Grades 1-6 correspond to school grades, as"
    " recorded in KANJIDIC.";
  std::string data3 = "Grades 7-9 are the rest of the Jouyou kanji,"
    " roughly divided into the same groups as Kanken levels 4 through 2.";
  std::string data4 = "Grade 10 is the new Jouyou kanji.";
  std::string data5 = "Grade 11 is the Jinmeiyou kanji.";

  m_textGameplay.setBackText("Back");
  m_textGameplay.addBlankLines(1);
  m_textGameplay.addTextItem(0, "GAMEPLAY");
  m_textGameplay.addBlankLines(2);
  m_textGameplay.addTextItem(0, gp1);
  m_textGameplay.addBlankLines(1);
  m_textGameplay.addTextItem(0, gp2);
  m_textGameplay.addBlankLines(1);
  m_textGameplay.addTextItem(0, gp3);

  m_textControls.setBackText("Back");
  m_textControls.addBlankLines(1);
  m_textControls.addTextItem(0, "CONTROLS");
  m_textControls.addBlankLines(2);
  m_textControls.addTextItem(0, "Move cursor:");
  m_textControls.addBlankLines(1);
  m_textControls.addTextItem(2, "Use left and right on the direction pad,"
			     " or drag with the stylus.");
  m_textControls.addBlankLines(2);
  m_textControls.addTextItem(0, "Fire missile:");
  m_textControls.addBlankLines(1);
  m_textControls.addTextItem(2, "Press A or "
			     "tap the space above the cursor.");
  m_textControls.addBlankLines(2);
  m_textControls.addTextItem(0, "Switch missile:");
  m_textControls.addBlankLines(1);
  m_textControls.addTextItem(2, "Press B or "
			     "tap the space diagonal to the cursor.");
  m_textControls.addBlankLines(2);
  m_textControls.addTextItem(0, "Pause:");
  m_textControls.addBlankLines(1);
  m_textControls.addTextItem(2, "Press Start.");

  m_textAcknowledgements.setBackText("Back");
  m_textAcknowledgements.addBlankLines(1);
  m_textAcknowledgements.addTextItem(0, "ACKNOWLEDGEMENTS");
  m_textAcknowledgements.addBlankLines(2);
  m_textAcknowledgements.addTextItem(0, ack_kcd);
  m_textAcknowledgements.addBlankLines(1);
  m_textAcknowledgements.addTextItem(0, ack_kanjidic);
  m_textAcknowledgements.addBlankLines(1);
  m_textAcknowledgements.addTextItem(0, ack_font);
  m_textAcknowledgements.addBlankLines(1);
  m_textAcknowledgements.addTextItem(0, ack_palib);
  m_textAcknowledgements.addBlankLines(1);
  m_textAcknowledgements.addTextItem(0, ack_ks);

  m_textData.setBackText("Back");
  m_textData.addBlankLines(1);
  m_textData.addTextItem(0, "DATA");
  m_textData.addBlankLines(2);
  m_textData.addTextItem(0, data1);
  m_textData.addBlankLines(1);
  m_textData.addTextItem(0, data2);
  m_textData.addBlankLines(1);
  m_textData.addTextItem(0, data3);
  m_textData.addBlankLines(1);
  m_textData.addTextItem(0, data4);
  m_textData.addBlankLines(1);
  m_textData.addTextItem(0, data5);
}

//------------------------------------------------------------------

void
MenuLoop::run()
{
  MenuPage state(Gameplay);
  if (m_bFirstLoad)
  {
    state = Title;
    m_bFirstLoad = false;
  }
  while (state != Game)
  {
    if (state == Gameplay)
    {
      m_menuGameplay.run();
      state = m_menuGameplay.exit();
    }
    else if (state == Data)
    {
      m_menuData.run();
      state = m_menuData.exit();
    }
    else if (state == Title)
    {
      m_menuTitle.run();
      state = m_menuTitle.exit();
    }
    else if (state == Guide)
    {
      m_menuGuide.run();
      state = m_menuGuide.exit();
    }
    else if (state == DeleteConfirm)
    {
      m_menuDeleteConfirm.run();
      state = m_menuDeleteConfirm.exit();
    }
    else if (state == LoadGame)
    {
      GameData::readNumberedFile(m_menuTitle.fileIndex(),
				 GameData::record);
      state = Gameplay;
    }
    else if (state == DeleteFile)
    {
      GameData::writeEmptyFile(m_menuTitle.fileIndex());
      state = Title;
    }
    else if (state == ViewData)
    {
      m_viewerData.run();
      state = Data;
    }
    else if (state == GuideGameplay)
    {
      m_textGameplay.run();
      state = Guide;
    }
    else if (state == GuideData)
    {
      m_textData.run();
      state = Guide;
    }
    else if (state == GuideControls)
    {
      m_textControls.run();
      state = Guide;
    }
    else if (state == GuideAcknowledgements)
    {
      m_textAcknowledgements.run();
      state = Guide;
    }
    else { state = Gameplay; }
  }
}

//-------------------------------------------------------------------

u32 MenuLoop::getSpeed() { return m_menuGameplay.getSpeed(); }

//-------------------------------------------------------------------

u8 MenuLoop::getMeteors() { return m_menuGameplay.getMeteors(); }

//-------------------------------------------------------------------

u8 MenuLoop::getFileIndex() const { return m_menuTitle.fileIndex(); }

//-------------------------------------------------------------------
